////////////////////////////////////////////////////////////////////////////
//	File:		Delegate2.h
//
//	Author:		Fabien Lebourg
//	E-mail:		fabien.lebourg@eole-soft.com
//
//	Implementation of Delegates with two arguments in C++.
//
//	This code may be used in compiled form in any way you desire. This
//	file may be redistributed unmodified by any means PROVIDING it is
//	not sold for profit without the authors written consent, and
//	providing that this notice and the authors name and all copyright
//	notices remains intact.
//
//	An email letting me know how you are using it would be nice as well.
//
//	This file is provided "as is" with no expressed or implied warranty.
//	The author accepts no liability for any damage/loss of business that
//	this product may cause.

#if !defined(_Delegate2_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
#define _Delegate2_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_

// Disable long templated symbole names alert in DebugMode !
#pragma warning( disable : 4786 )

#include <assert.h>
#include <vector>
#include <list>
#include <algorithm>
using namespace std;


#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


//*************************************************************************************
// For particular case (see doc)
//*************************************************************************************
#define MEMBER_FUNCTION_CALL2( CLASS, ARG1, ARG2 ) void (CLASS::* )( ARG1, ARG2 )
#define MEMBER_FUNCTION_CALL_TYPEDEF2( CLASS, ARG1, ARG2, NAME ) void (CLASS::* NAME)( ARG1, ARG2 )
//*************************************************************************************


//*************************************************************************************
// CDelegateImp2_Call
//*************************************************************************************
template<class ARG1, class ARG2>
class CDelegateImp2_Call
{
public:
   virtual void operator( ) ( ARG1&, ARG2& ) = 0;
   virtual bool Match( int pfn, void* pobj ) const = 0;
};
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp2_Call_Static
//*************************************************************************************
template<class ARG1, class ARG2>
class CDelegateImp2_Call_Static : public CDelegateImp2_Call<ARG1, ARG2>
{
public:
   typedef void (*PFN)( ARG1, ARG2 );
   CDelegateImp2_Call_Static( PFN pfn ) : m_pfn( pfn ) 
   { 
      assert( pfn ); 
   }
   virtual void operator( ) ( ARG1& arg1, ARG2& arg2 ) 
   { 
      (*m_pfn)( arg1, arg2 ); 
   }
   virtual bool Match( int pfn, void* pobj ) const
   {
      return (!pobj) && (m_pfn == (PFN) pfn);
   }
protected:
   PFN   m_pfn;
};
//*************************************************************************************


//*************************************************************************************
//class CDelegateImp2_Call_Member
//*************************************************************************************
template<class CLASS, class ARG1, class ARG2>
class CDelegateImp2_Call_Member : public CDelegateImp2_Call<ARG1, ARG2>
{
public:
   typedef void (CLASS::* PFN)( ARG1, ARG2 );

   CDelegateImp2_Call_Member( CLASS& object, PFN pfn, int iFncID ) 
   :  m_object( object ), 
      m_pfn( pfn ), 
      m_iFncID( iFncID )
   { 
      assert( pfn ); 
   }
   virtual void operator( ) ( ARG1& arg1, ARG2& arg2 ) 
   { 
      (m_object.*m_pfn)( arg1, arg2 ); 
   }
   virtual bool Match( int iFncID, void* pobj ) const
   {
      return (&m_object == (CLASS*)pobj) && iFncID == m_iFncID;
   }
protected:
   PFN      m_pfn;
   CLASS&   m_object;
   int      m_iFncID;
};
//*************************************************************************************


//*************************************************************************************
// class CClassInfo2
//*************************************************************************************
//
// Associate an ID and a member function address
//
template<class CLASS, class ARG1, class ARG2>
class CClassInfo2
{
public:
   typedef  void (CLASS::* PFN)( ARG1, ARG2 );
   
   static int GetFunctionID( PFN pfn )
   {
      vector<PFN>::iterator it = find( m_vecFns.begin( ), m_vecFns.end( ), pfn );
      if( it == m_vecFns.end( ) )
      {
         m_vecFns.push_back( pfn );
         return m_vecFns.size( ) - 1;
      }
      else
      {
         return m_vecFns.begin( ) - it;
      }
   };

protected:

   static vector<PFN> m_vecFns;

};
//*************************************************************************************
template<class CLASS, class ARG1, class ARG2>
vector< CClassInfo2<CLASS, ARG1, ARG2>::PFN > CClassInfo2<CLASS, ARG1, ARG2>::m_vecFns;
//*************************************************************************************


//*************************************************************************************
// class CDelegateImp2
//*************************************************************************************
// 
// Delegate with two arguments
//
template<class CS, class ARG1, class ARG2>
class CDelegateImp2  
{
public:

// --- typedefs

   typedef CDelegateImp2_Call< ARG1, ARG2 >   CALL;
   typedef list< CALL* >            CALLS;
   typedef CALLS::const_iterator    CALLS_CIT;
   typedef CALLS::iterator          CALLS_IT;

   
// --- Construction & Destruction

   CDelegateImp2( ) 
   { 
   }
   
   virtual ~CDelegateImp2( ) 
   { 
      Clear( ); 
   }


// --- Operations

	//-----------------------------------------------------------------
   // Execution
	//-----------------------------------------------------------------
   void  operator( ) ( ARG1 arg1, ARG2 arg2 ) const
   {
      m_cs.Lock( );
      
      CCallExecute<ARG1, ARG2> exec( arg1, arg2 );
      for_each( m_calls.begin( ), m_calls.end( ), exec );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


   //-----------------------------------------------------------------
   // Assign
	//-----------------------------------------------------------------
   const CDelegateImp2& operator = ( const CDelegateImp2& other )
   {
      m_cs.Lock( );
      m_calls = other.m_calls;
      m_cs.Unlock( );
      return *this;
   }
   //-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Add member function address
	//-----------------------------------------------------------------   
   template<class OBJECT, class CLASS>
   void  Add( const OBJECT& object, void (CLASS::* fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo2<CLASS, ARG1, ARG2>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp2_Call<ARG1, ARG2>* pcall = new CDelegateImp2_Call_Member<CLASS, ARG1, ARG2>( *(OBJECT*)&object, fnc, iFnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Add static function address
	//-----------------------------------------------------------------
   void  Add( void (*fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder);
      if( it != m_calls.end( ) )
         return;

      CDelegateImp2_Call<ARG1, ARG2>* pcall = new CDelegateImp2_Call_Static<ARG1, ARG2>( fnc );
      m_calls.push_back( pcall );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

   
	//-----------------------------------------------------------------   
   // Remove static function address
	//-----------------------------------------------------------------
   void  Remove( void (*fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
      {
         delete *it;
         m_calls.erase( it );
      }

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Remove member function address
	//-----------------------------------------------------------------
   template<class OBJECT, class CLASS>
   void  Remove( const OBJECT& object, void (CLASS::* fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo2<CLASS, ARG1, ARG2>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      if( it != m_calls.end( ) )
         m_calls.erase( it );

      m_cs.Unlock( );
   }      
	//-----------------------------------------------------------------


   //-----------------------------------------------------------------   
   // Has static function address ?
	//-----------------------------------------------------------------
   bool Has( void (*fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      CCallFinder finder( (long) fnc, (void*) 0 );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Has member function address ?
	//-----------------------------------------------------------------
   template<class CLASS>
   bool Has( const CLASS& object, void (CLASS::* fnc)( ARG1, ARG2 ) )
   {
      m_cs.Lock( );

      int iFnc = CClassInfo2<CLASS, ARG1, ARG2>::GetFunctionID( fnc );

      CCallFinder finder( iFnc, (void*) &object );
      CALLS_IT it = find_if(m_calls.begin( ), m_calls.end( ), finder );
      bool bret = (it != m_calls.end( ));

      m_cs.Unlock( );
      return bret;
   }      
	//-----------------------------------------------------------------


	//-----------------------------------------------------------------   
   // Clear Delegate (remove all references)
	//-----------------------------------------------------------------
   void  Clear( )
   {
      m_cs.Lock( );

      for_each( m_calls.begin( ), m_calls.end( ), DeleteCall );
      m_calls.erase( m_calls.begin( ), m_calls.end( ) );

      m_cs.Unlock( );
   }
	//-----------------------------------------------------------------

protected:

// --- Data Members

   CALLS       m_calls;
   mutable CS  m_cs;

// --- Algo Func

	//-----------------------------------------------------------------   
   struct CCallFinder
   {
      CCallFinder( int pfn, void* pobj ) : m_pfn( pfn ), m_pobj( pobj ) { }
      int operator( )( CDelegateImp2_Call < ARG1, ARG2 >* pcall )
      {
         return pcall->Match( m_pfn, m_pobj );
      }
   protected:
      int     m_pfn;
      void*    m_pobj;
   };
	//-----------------------------------------------------------------
   template<class ARG1, class ARG2>
   struct CCallExecute
   {
      CCallExecute( ARG1& arg1, ARG2& arg2 ) : m_arg1( arg1 ), m_arg2( arg2 ) { }
      void operator( )( CDelegateImp2_Call < ARG1, ARG2 >* pcall )
      {
         assert( pcall );
         (*pcall)( m_arg1, m_arg2 );
      }
   protected:
      ARG1& m_arg1;
      ARG1& m_arg2;
   };
	//-----------------------------------------------------------------
   static void DeleteCall( CALL* pCall )
   {
      assert( pCall );
      delete pCall;
   }
	//-----------------------------------------------------------------   

};
//*************************************************************************************

#endif // !defined(_Delegate2_H__8256AA13_8F1A_445B_A1B7_D8CCB52F97D5__INCLUDED_)
